from .dictionaries import DictObject, apply_dict_object, DictionaryView, TwoWayDictionary, \
    DirtyDict, KeyAwareDefaultDict, ExpiringEntryDict, SelfCleaningDefaultDict, \
    CacheDict
from .hashable_objects import HashableWrapper
from .heaps import Heap, SetHeap, TimeBasedHeap, TimeBasedSetHeap
from .immutable import Immutable, frozendict
from .mixins import OmniHashableMixin, ReprableMixin, StrEqHashableMixin, ComparableIntEnum, \
    HashableIntEnum, ComparableAndHashableBy
from .proxy import Proxy
from .ranking import Ranking
from .singleton import Singleton, SingletonWithRegardsTo
from .sorted_list import SortedList, SliceableDeque
from .typednamedtuple import typednamedtuple

__all__ = [
    'CacheDict',
    'KeyAwareDefaultDict',
    'Proxy',
    'ReprableMixin',
    'StrEqHashableMixin',
    'ComparableIntEnum',
    'ComparableAndHashableBy',
    'HashableIntEnum',
    'DirtyDict',
    'SortedList',
    'SelfCleaningDefaultDict',
    'SliceableDeque',
    'Ranking',
    'ExpiringEntryDict',
    'TwoWayDictionary',
    'typednamedtuple',
    'HashableWrapper',
    'DictionaryView',
    'frozendict',
    'OmniHashableMixin',
    'Singleton',
    'SingletonWithRegardsTo',
    'DictObject',
    'apply_dict_object',
    'Immutable',
    'SetHeap', 'TimeBasedHeap', 'TimeBasedSetHeap', 'Heap'
]



import time
import typing as tp
from satella.coding.structures import CacheDict
from .. import Metric
from ..metric_types.measurable_mixin import MeasurableMixin


class MetrifiedCacheDict(CacheDict):
    """
    A CacheDict with metrics!

    :param cache_hits: a counter metric that will be updated with +1 each time there's a cache hit
    :param cache_miss: a counter metric that will be updated with +1 each time there's a cache miss
    :param refreshes: a metric that will be updated with +1 each time there's a cache refresh
    :param how_long_refresh_takes: a metric that will be ticked with time value_getter took
    """
    def __init__(self, stale_interval, expiration_interval, value_getter,
                 value_getter_executor=None, cache_failures_interval=None,
                 time_getter=time.monotonic,
                 default_value_factory=None,
                 cache_hits: tp.Optional[Metric] = None,
                 cache_miss: tp.Optional[Metric] = None,
                 refreshes: tp.Optional[Metric] = None,
                 how_long_refresh_takes: tp.Optional[MeasurableMixin] = None):
        if refreshes:
            old_value_getter = value_getter

            def value_getter_replacement(item):
                try:
                    return old_value_getter(item)
                finally:
                    if self.refreshes:
                        self.refreshes.runtime(+1)

            value_getter = value_getter_replacement

        if how_long_refresh_takes:
            value_getter = how_long_refresh_takes.measure(value_getter=time_getter)(value_getter)

        super().__init__(stale_interval, expiration_interval, value_getter,
                         value_getter_executor, cache_failures_interval, time_getter,
                         default_value_factory)
        self.cache_hits = cache_hits
        self.cache_miss = cache_miss
        self.refreshes = refreshes
        self.how_long_refresh_takes = how_long_refresh_takes

    def __getitem__(self, item):
        if self.has_info_about(item):
            if self.cache_hits:
                self.cache_hits.runtime(+1)
        else:
            if self.cache_miss:
                self.cache_miss.runtime(+1)
        return super().__getitem__(item)



"""
Just useful objects to make your coding nicer every day
"""

from .algos import merge_dicts
from .concurrent import Monitor, RMonitor
from .decorators import precondition, short_none, has_keys, \
    wraps, chain_functions, postcondition, queue_get, auto_adapt_to_methods, \
    attach_arguments, for_argument
from .deleters import ListDeleter, DictDeleter
from .fun_static import static_var
from .iterators import hint_with_length, SelfClosingGenerator, exhaust, chain
from .metaclasses import metaclass_maker, wrap_with, dont_wrap, wrap_property, DocsFromParent
from .misc import update_if_not_none
from .recast_exceptions import rethrow_as, silence_excs, catch_exception, log_exceptions, \
    raises_exception

__all__ = [
    'update_if_not_none',
    'hint_with_length', 'SelfClosingGenerator', 'exhaust', 'chain',
    'Monitor', 'RMonitor', 'merge_dicts',
    'short_none', 'has_keys',
    'precondition', 'postcondition', 'queue_get',
    'rethrow_as', 'silence_excs', 'raises_exception',
    'static_var', 'metaclass_maker',
    'catch_exception', 'wraps', 'wrap_with', 'dont_wrap', 'wrap_property', 'log_exceptions',
    'chain_functions',
    'ListDeleter', 'DictDeleter', 'for_argument', 'attach_arguments', 'auto_adapt_to_methods',
]



import collections
import copy
import heapq
import typing as tp

from satella.coding.decorators import wraps

T = tp.TypeVar('T')


def _extras_to_one(fun):
    @wraps(fun)
    def inner(self, a, *args, **kwargs):
        return fun(self, ((a,) + args) if len(args) > 0 else a, **kwargs)

    return inner


class Heap(collections.UserList, tp.Generic[T]):
    """
    Sane heap as object - not like heapq.

    Goes from lowest-to-highest (first popped is smallest).
    Standard Python comparision rules apply.

    Not thread-safe
    """

    def __init__(self, from_list: tp.Optional[tp.Iterable[T]] = None):
        super().__init__(from_list)
        heapq.heapify(self.data)

    def push_many(self, items: tp.Iterable[T]) -> None:
        for item in items:
            self.push(item)

    def pop_item(self, item: T) -> T:
        """
        Pop an item off the heap, maintaining the heap invariant

        :raise ValueError: element not found
        """
        self.data.remove(item)  # raises: ValueError
        heapq.heapify(self.data)
        return item

    @_extras_to_one
    def push(self, item: T) -> None:
        """
        Use it like:

        >>> heap.push(3)

        or:

        >>> heap.push(4, myobject)
        """
        heapq.heappush(self.data, item)

    def __deepcopy__(self, memo) -> 'Heap':
        return self.__class__(copy.deepcopy(self.data, memo))

    def __copy__(self) -> 'Heap':
        return self.__class__(copy.copy(self.data))

    def __iter__(self) -> tp.Iterator[T]:
        return self.data.__iter__()

    def pop(self) -> T:
        """
        Return smallest element of the heap.

        :raises IndexError: on empty heap
        """
        return heapq.heappop(self.data)

    def filter_map(self, filter_fun: tp.Optional[tp.Callable[[T], bool]] = None,
                   map_fun: tp.Optional[tp.Callable[[T], tp.Any]] = None):
        """
        Get only items that return True when condition(item) is True. Apply a
         transform: item' = item(condition) on
        the rest. Maintain heap invariant.
        """
        heap = filter(filter_fun, self.data) if filter_fun else self.data
        heap = map(map_fun, heap) if map_fun else heap
        heap = list(heap) if not isinstance(heap, list) else heap
        self.data = heap
        heapq.heapify(self.data)

    def __bool__(self) -> bool:
        """
        Is this empty?
        """
        return len(self.data) > 0

    def iter_ascending(self) -> tp.Iterable[T]:
        """
        Return an iterator returning all elements in this heap sorted ascending.
        State of the heap is not changed
        """
        heap = copy.copy(self.data)
        while heap:
            yield heapq.heappop(heap)

    def iter_descending(self) -> tp.Iterable[T]:
        """
        Return an iterator returning all elements in this heap sorted descending.
        State of the heap is not changed.

        This loads all elements of the heap into memory at once, so be careful.
        """
        return reversed(list(self.iter_ascending()))

    def __eq__(self, other: 'Heap') -> bool:
        return self.data == other.data

    def __len__(self) -> int:
        return len(self.data)

    def __str__(self) -> str:
        return '<satella.coding.Heap: %s elements>' % (len(self, ))

    def __repr__(self) -> str:
        return u'<satella.coding.Heap>'

    def __contains__(self, item: T) -> bool:
        return item in self.data


class SetHeap(Heap):
    """
    A heap with additional invariant that no two elements are the same.

    Note that elements you insert in this must be eq-able and hashable, ie. you can put them in a dict.

    Optimized for fast insertions and fast __contains__

    #notthreadsafe
    """

    def __init__(self, from_list: tp.Optional[tp.Iterable[T]] = None):
        super().__init__(from_list=from_list)
        self.set = set(self.data)

    def push(self, item: T):
        if item not in self.set:
            super().push(item)
            self.set.add(item)

    def pop(self) -> T:
        item = super().pop()
        self.set.remove(item)
        return item

    def __contains__(self, item: T) -> bool:
        return item in self.set

    def filter_map(self, filter_fun: tp.Optional[tp.Callable[[T], bool]] = None,
                   map_fun: tp.Optional[tp.Callable[[T], tp.Any]] = None):
        super().filter_map(filter_fun=filter_fun, map_fun=map_fun)
        self.set = set(self.data)



import copy
import typing as tp

K, V, T = tp.TypeVar('K'), tp.TypeVar('V'), tp.TypeVar('T')


class DirtyDict(tp.MutableMapping[K, V]):
    """
    A dictionary that has also a flag called .dirty that sets to True if the dictionary has been
    changed since that flag was last cleared.

    Setting the dict with the value that it already has doesn't count as dirtying it.
    Note that such changes will not be registered in the dict!

    All arguments and kwargs will be passed to dict's constructor.
    """

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> tp.Iterator[K]:
        return iter(self.data)

    def __getitem__(self, k: K) -> V:
        return self.data[k]

    def __init__(self, *args, **kwargs):
        self.data = dict(*args, **kwargs)  # type: tp.Dict[K, V]
        self.dirty = False

    def __copy__(self) -> 'DirtyDict':
        dd = DirtyDict(self.data.copy())
        dd.dirty = self.dirty
        return dd

    def __setitem__(self, key: K, value: V) -> None:
        if key in self.data:
            if self.data[key] == value:
                return
        self.data[key] = value
        self.dirty = True

    def __delitem__(self, key: K) -> None:
        del self.data[key]
        self.dirty = True

    def __contains__(self, item: K) -> bool:
        return item in self.data

    def clear_dirty(self) -> None:
        """Clears the dirty flag"""
        self.dirty = False

    def __bool__(self) -> bool:
        return bool(self.data)

    def swap_and_clear_dirty(self) -> tp.Dict[K, V]:
        """
        Returns this data, clears self and sets dirty to False

        After this is called, this dict will be considered empty.

        :return: a plain, normal Python dictionary is returned
        """
        a = self.data
        self.data = {}
        self.dirty = False
        return a

    def copy_and_clear_dirty(self) -> tp.Dict[K, V]:
        """
        Returns a copy of this data and sets dirty to False

        :return: a plain, normal Python dictionary is returned
        """
        a = self.data.copy()
        self.dirty = False
        return a


class KeyAwareDefaultDict(tp.MutableMapping[K, V]):
    """
    A defaultdict whose factory function accepts the key to provide a default value for the key

    :param factory_function: a callable that accepts a single argument, a key, for which it is
        to provide a default value
    """

    def __len__(self) -> int:
        return len(self.dict)

    def __iter__(self) -> tp.Iterator[K]:
        return iter(self.dict)

    def __init__(self, factory_function: tp.Callable[[K], V], *args, **kwargs):
        self.dict = dict(*args, **kwargs)
        self.factory_function = factory_function

    def __getitem__(self, item) -> V:
        if item in self.dict:
            return self.dict[item]
        else:
            self.dict[item] = self.factory_function(item)
            return self.dict[item]

    def __setitem__(self, key, value) -> None:
        self.dict[key] = value

    def __delitem__(self, key) -> None:
        del self.dict[key]


class TwoWayDictionary(tp.MutableMapping[K, V]):
    """
    A dictionary that keeps also a reverse_data mapping, allowing to look up keys by values.

    Not thread-safe.

    Example usage:

    >>> twd = TwoWayDictionary()
    >>> twd[2] = 3
    >>> self.assertEqual(twd.reverse[3], 2)

    When you're done using a given TwoWayDictionary, please call .done(). This will make it easier for the GC to collect
    the dictionaries.

    You can also use the context manager to make the TwoWayDictionary clean up itself, eg.

    >>> with TwoWayDictionary() as twd:
    >>>     ...
    >>> # at this point twd is .done()

    :param data: data to generate the dict from
    :raises ValueError: on being given data from which it is impossible to construct a reverse
        mapping (ie. same value appears at least twice)
    """
    __slots__ = ('data', 'reverse_data', '_reverse')

    def done(self) -> None:
        """
        Called when the user is done using given TwoWayDictionary.

        Internally this will break the reference cycle, and enable Python GC to collect the objects.
        """
        self.reverse.reverse = None
        self.reverse = None

    def __init__(self, data=None, _is_reverse: bool = False):
        if not _is_reverse:
            self.data = dict(data or [])
            self.reverse_data = {v: k for k, v in self.data.items()}
            if len(self.reverse_data) != len(self.data):
                raise ValueError('Value repeats itself, invalid data!')

            self._reverse = TwoWayDictionary(_is_reverse=True)
            self._reverse.data = self.reverse_data
            self._reverse.reverse_data = self.data
            self._reverse._reverse = self

    def __enter__(self) -> 'TwoWayDictionary':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        self.done()
        return False

    def __getitem__(self, item: K) -> V:
        return self.data[item]

    def __len__(self) -> int:
        return len(self.data)

    def __setitem__(self, key: K, value: V) -> None:
        if value in self.reverse_data:
            raise ValueError('This value is already mapped to something!')

        try:
            prev_val = self.data[key]
        except KeyError:
            pass
        else:
            del self.reverse_data[prev_val]

        self.data[key] = value
        self.reverse_data[value] = key

    def __delitem__(self, key: K) -> None:
        value = self.data[key]
        del self.data[key]
        del self.reverse_data[value]

    def __iter__(self) -> tp.Iterator[K]:
        return iter(self.data)

    @property
    def reverse(self) -> tp.MutableMapping[V, K]:
        """
        Return a reverse mapping. Reverse mapping is updated as soon as an operation is done.
        """
        return self._reverse


class DictionaryView(tp.MutableMapping[K, V]):
    """
    A view on a multiple dictionaries. If key isn't found in the first dictionary, it is looked up
    in another. Use like:

    >>> dv = DictionaryView({1:2, 3:4}, {4: 5, 6: 7})
    >>> assert dv[4] == 5
    >>> del dv[1]
    >>> assertRaises(KeyError, lambda: dv.__delitem__(1))

    :param master_dict: First dictionary to look up. Entries made via __setitem__ will be put here.
    :param rest_of_dicts: Remaining dictionaries
    :param propagate_deletes: Whether to delete given key from the first dictionary that it is
        found. Otherwise it will be only deleted from the master_dict. Also, if this is set to
        False, on deletion, if the key isn't found in master dictionary, deletion will KeyError.
    :param assign_to_same_dict: whether updates done by __setitem__ should be written to the
        dictionary that contains that key. If not, all updates will be stored in master_dict. If
        this is True, updates made to keys that are not in this dictionary will go to master_dict.
    """
    __slots__ = ('assign_to_same_dict', 'master_dict', 'dictionaries', 'propagate_deletes')

    def __copy__(self) -> 'DictionaryView':
        return DictionaryView(*copy.copy(self.dictionaries))

    def __deepcopy__(self, memo) -> 'DictionaryView':
        return DictionaryView(*copy.deepcopy(self.dictionaries, memo))

    def __init__(self, master_dict: tp.Dict[K, V], *rest_of_dicts: tp.Dict[K, V],
                 propagate_deletes: bool = True,
                 assign_to_same_dict: bool = True):
        self.assign_to_same_dict = assign_to_same_dict
        self.master_dict = master_dict
        self.dictionaries = [master_dict, *rest_of_dicts]
        self.propagate_deletes = propagate_deletes

    def __contains__(self, item: K) -> bool:
        for dictionary in self.dictionaries:
            if item in dictionary:
                return True
        return False

    def __iter__(self) -> tp.Iterator[K]:
        seen_already = set()
        for dictionary in self.dictionaries:
            for key in dictionary:
                if key not in seen_already:
                    yield key
                    seen_already.add(key)

    def __len__(self) -> int:
        seen_already = set()
        i = 0
        for dictionary in self.dictionaries:
            for key in dictionary:
                if key not in seen_already:
                    i += 1
                    seen_already.add(key)
        return i

    def __getitem__(self, item: K) -> V:
        for dictionary in self.dictionaries:
            if item in dictionary:
                return dictionary[item]
        raise KeyError('Key not found')

    def __setitem__(self, key: K, value: V) -> None:
        if self.assign_to_same_dict:
            for dictionary in self.dictionaries:
                if key in dictionary:
                    dictionary[key] = value
                    return
        self.master_dict[key] = value

    def __delitem__(self, key: K) -> None:
        if self.propagate_deletes:
            for dictionary in self.dictionaries:
                if key in dictionary:
                    del dictionary[key]
                    return
            raise KeyError('Key not found')
        else:
            del self.master_dict[key]



import signal
import sys
import types
import typing as tp

try:
    # noinspection PyProtectedMember
    SIG_TYPE = signal._SIG
except AttributeError:
    SIG_TYPE = int  # we are running on Windows


# noinspection PyUnusedLocal
def dump_frames_on(sig_no: SIG_TYPE, stack_frame: types.FrameType, output: tp.TextIO):
    from satella.instrumentation import Traceback

    output.write("Stack frame dump requested in response to signal %s\n" % (sig_no,))
    # noinspection PyProtectedMember
    for frame_no, frame in sys._current_frames().items():
        output.write("For stack frame %s" % (frame_no,))
        tb = Traceback(frame)
        tb.pretty_print(output=output)
    output.write("End of stack frame dump\n")


def install_dump_frames_on(signal_number: SIG_TYPE, output: tp.TextIO = sys.stderr):
    """
    Instruct Python to dump all frames onto output, along with their local variables
    upon receiving given signal
    """
    signal.signal(signal_number,
                  lambda sig_no, stack_frame: dump_frames_on(sig_no, stack_frame, output))



import collections
import random
import typing as tp

__all__ = ['stringify', 'split_shuffle_and_join', 'one_tuple']


def _stringify_none(str_none, stringifier):
    if str_none:
        return stringifier(None)
    return None


T = tp.TypeVar('T')


def one_tuple(x: tp.Iterable[T]) -> tp.Iterator[tp.Tuple[T]]:
    """
    Change a sequence of iterables into a sequence that displays each element as
    a part of one-element tuple. Essentially syntactic sugar for:

    >>> for y in x:
    >>>     yield y,

    :param x: sequence to tupleify
    :return: a iterator of one-element tuples
    """
    for y in x:
        yield y,


def split_shuffle_and_join(entries: tp.List[T],
                           whether_to_shuffle: tp.Callable[[T], bool] = lambda x: True,
                           not_shuffled_to_front: bool = True) -> tp.List[T]:
    """
    Split elements in entries into two groups - one group, called True, is the one for which
    whether_to_shuffle(elem) is True, the other is False.

    Shuffle the group True.

    If not_shuffled_to_front, elements in the group False will go at the beginning of the returned
    list, after which will go elements shuffled. If it's False, the not-shuffled elements will be
    at the back of the list.

    Order of the not shuffled elements will be preserved.

    :param entries: list of elements
    :param whether_to_shuffle: a decider to which group does given element belong?
    :param not_shuffled_to_front: if True then not shuffled elements will be put before shuffled,
        else the not shuffled elements will be at the back of the list
    :return: list altered to specification
    """

    not_shuffled, shuffled = [], []
    for elem in entries:
        (shuffled if whether_to_shuffle(elem) else not_shuffled).append(elem)

    random.shuffle(shuffled)

    if not_shuffled_to_front:
        return not_shuffled + shuffled
    else:
        return shuffled + not_shuffled


def stringify(obj: tp.Union[tp.Any], stringifier: tp.Callable[[tp.Any], str] = str,
              recursively: bool = False,
              str_none: bool = False) -> tp.Dict[str, str]:
    """
    Stringify all object:

    ie. if a dict, put every item and key (if a dict is given) through stringify.
        if a list, put every item through stringify
        else just call stringify on it.

        Note that if you use recursively, then dicts and lists are allowed to be valid elements of the returned
        representation!

    :param obj: a list or a dict
    :param stringifier: function that accepts any arguments and returns a string representation
    :param recursively: whether to recursively stringify elements, ie. stringify will be called on all the children
    :param str_none: whether to return None if given a None. If True, "None" will be returned instead
    :return: stringified object
    """
    if isinstance(obj, collections.abc.Mapping):
        make_str = (lambda obj2: stringify(obj2, stringifier, True, str_none)) if recursively else \
            stringifier
        return {make_str(k): make_str(v) for k, v in obj.items()}
    elif isinstance(obj, collections.abc.Sequence):
        make_str = (lambda obj2: stringify(obj2, stringifier, True, str_none)) if recursively else \
            stringifier
        return [make_str(v) for v in obj]
    elif obj is None:
        return _stringify_none(str_none, stringifier)
    else:
        return stringifier(obj)
